/*
Note: 
	IIRC, These functions are mine. I just created them from Raymond's samples.
*/
#define STRICT_TYPED_ITEMIDS
#include <windows.h>
#include "ShellCOMFunctions.h"
#include <shlobj.h>
#include <assert.h>
#include <shlwapi.h>
#include <comdef.h>
#include <tchar.h>
#include <atlconv.h>

//Wrapper for IShellWindows::Item() : Get the shell window's IDispatch by index.
EXTERN_C HRESULT IShellWindows_Item_int(IShellWindows *This, INT32 index, IDispatch **ppDispFolder)
{
	VARIANT vIndex;
	VariantInit(&vIndex);
	V_VT(&vIndex) = VT_I4;
	V_I4(&vIndex) = index;
	HRESULT hr = This->Item(vIndex, ppDispFolder);
	VariantClear(&vIndex);
	return hr;
}

//Create a ShellWindows object, and call enumProc on its contents.
EXTERN_C HRESULT EnumShellWindows(INT_PTR userData, SHLWNDENUMPROC enumProc)
{
	if(enumProc==NULL)
		return E_POINTER;
	
	HRESULT hr;
	IShellWindowsPtr spWindows = NULL;
	hr = CoCreateInstance(CLSID_ShellWindows, NULL, CLSCTX_INPROC_SERVER|CLSCTX_LOCAL_SERVER, IID_PPV_ARGS(&spWindows));
	if(FAILED(hr))
	{
		switch(hr)
		{
		case REGDB_E_CLASSNOTREG: //If this occurs, it's very unexpected, but identifiable.
		case 0x804F4C45:
			return hr;
		default:
			return E_UNEXPECTED;
		}
	}
	assert(spWindows!=NULL);
	
	hr = S_OK;
	IDispatchPtr spDispWindow=NULL;
	for(UINT index=0 ; hr==S_OK && IShellWindows_Item_int(spWindows, index, &spDispWindow)==S_OK ; index++)
	{
		HRESULT hrProc = enumProc(userData, spDispWindow);
		if(FAILED(hrProc))
			hr = E_UNEXPECTED;
		if(hrProc==S_FALSE)
			hr = S_FALSE;
		
		spDispWindow.Release();
	}
	
	return hr;
}

#pragma region Functions to find a shell window from any criterion
template<class T>
struct st_findShellWindow
{
	typedef BOOL (CALLBACK * FINDPROC)(T userData, IDispatch *pDispShellWindow);

	T userData;
	FINDPROC findProc;
	IDispatchPtr spDispFoundShellWindow;
};

template<class T>
static HRESULT CALLBACK FindShellWindowEnumProc(INT_PTR userData, IDispatch *pDispShellWindow)
{
	HRESULT hr = S_OK; //Continue enumeration
	struct st_findShellWindow<T> * pData = reinterpret_cast<struct st_findShellWindow<T>*>(userData);
	bool isMatchingWindow = false;

	//Compare windows
	{
		if(pData->findProc(pData->userData, pDispShellWindow))
			isMatchingWindow = true;
	}//local variable

	if(SUCCEEDED(hr) && isMatchingWindow)
	{
		pData->spDispFoundShellWindow.Attach(pDispShellWindow, true);
		hr = S_FALSE; //Stop enumeration
	}
	return hr;
}

//Calls EnumShellWindows() with FindShellWindowEnumProc.
template<class T>
static HRESULT FindShellWindowT(T userData, typename st_findShellWindow<T>::FINDPROC findProc, IDispatch **ppDispShellWindow)
{
	if(ppDispShellWindow==NULL)
		return E_POINTER;
	*ppDispShellWindow=NULL;
	
	struct st_findShellWindow<T> data = {userData, findProc, NULL};
	HRESULT hr = EnumShellWindows(reinterpret_cast<INT_PTR>(&data), FindShellWindowEnumProc<T>);
	//Debug
	if(hr == 0x804F4C45)
		return hr;
	if(FAILED(hr))
		return E_UNEXPECTED;
	
	if(hr==S_FALSE)
	{
		//Enumeration was interrupted: found
		assert(data.spDispFoundShellWindow!=NULL);
		//Transfer ownership of the interface
		*ppDispShellWindow = data.spDispFoundShellWindow.Detach();
		return S_OK;
	}
	else
	{
		//Enumeration was not interrupted: not found.
		assert(data.spDispFoundShellWindow==NULL);
		*ppDispShellWindow=NULL;
		return S_FALSE;
	}
}
EXTERN_C HRESULT FindShellWindow(INT_PTR userData, SHLWNDFINDPROC findProc, IDispatch **ppDispShellWindow)
{
	return FindShellWindowT<INT_PTR>(userData, findProc, ppDispShellWindow);
}
#pragma endregion
#pragma region Functions to find a shell window from its HWND
struct st_findShellWindowByHandle
{
	HWND hWndToFind;
};

static BOOL CALLBACK FindShellWindowByHandleFindProc(struct st_findShellWindowByHandle * pData, IDispatch *pDispShellWindow)
{
	HRESULT hr = S_OK; //Continue enumeration
	BOOL ret = false;

	//Use window's IWebBrowserApp to get its HWND and compare
	{
		IWebBrowserAppPtr spBrowser = NULL;
		if(SUCCEEDED(pDispShellWindow->QueryInterface(&spBrowser)))
		{
			assert(spBrowser!=NULL);
			SHANDLE_PTR hpHwnd = 0;
			if(FAILED(spBrowser->get_HWND(&hpHwnd)))
				hr = E_UNEXPECTED;
			else
			{
				hr = S_OK;
				HWND hWndFound = reinterpret_cast<HWND>(hpHwnd);
				//Got the shell window's browser's HWND.

				//Compare.
				if(pData->hWndToFind == hWndFound)
					ret = true;
			}
		}
	}//local variable
	return ret;
}

//Calls EnumShellWindows() with FindShellWindowEnumProc.
EXTERN_C HRESULT FindShellWindowByHandle(HWND hWndToFind, IDispatch **ppDispShellWindow)
{
	struct st_findShellWindowByHandle data = {hWndToFind};
	return FindShellWindowT(&data, FindShellWindowByHandleFindProc, ppDispShellWindow);
}
#pragma endregion
#pragma region Functions to find a shell window from its PIDL
struct st_findShellWindowByPIDL
{
	PCIDLIST_ABSOLUTE pidl;
};

static BOOL CALLBACK FindShellWindowByPIDLFindProc(struct st_findShellWindowByPIDL * pData, IDispatch *pDispShellWindow)
{
	HRESULT hr = S_OK; //Continue enumeration
	BOOL ret = false;

	{
		//IFolderViewPtr spView;
		_com_ptr_t< _com_IIID<IFolderView, &__uuidof(IFolderView)> > spView;
		LPCTSTR sczErr;
		hr = GetShellWindowFolderView(pDispShellWindow, IID_PPV_ARGS(&spView), &sczErr);
		if(SUCCEEDED(hr))
		{
			IPersistFolder2Ptr spPersist;
			if (SUCCEEDED( spView->GetFolder(IID_PPV_ARGS(&spPersist)) ))
			{
				PIDLIST_ABSOLUTE pidlFolder;
				if (SUCCEEDED(spPersist->GetCurFolder(&pidlFolder)))
				{
					ret = ILIsEqual(pidlFolder, pData->pidl);
					
					CoTaskMemFree(pidlFolder);
				}
			}
		}
	}//local variable
	return ret;
}

//Calls EnumShellWindows() with FindShellWindowEnumProc.
EXTERN_C HRESULT FindShellWindowByPIDL(PCIDLIST_ABSOLUTE pidlToFind, IDispatch **ppDispShellWindow)
{
	struct st_findShellWindowByPIDL data = {pidlToFind};
	return FindShellWindowT(&data, FindShellWindowByPIDLFindProc, ppDispShellWindow);
}
#pragma endregion

//Get the requested interface on a given shell window's shell view.
EXTERN_C HRESULT GetShellWindowFolderView(IDispatch *pDispShellWindow, REFIID riid, void **ppView, LPCTSTR *pErr)
{
	if(pDispShellWindow==NULL || ppView==NULL)
		return E_POINTER;
	*ppView = NULL;
	
	//Get the shell windows's top-level shell browser's active shell view.
	HRESULT hr = E_UNEXPECTED;
	IServiceProviderPtr spProv = NULL;
	if(SUCCEEDED(hr = pDispShellWindow->QueryInterface(&spProv)))
	{
		assert(spProv!=NULL);
		IShellBrowserPtr spBrowser = NULL;
		if(SUCCEEDED(hr = spProv->QueryService(SID_STopLevelBrowser, &spBrowser)))
		{
			IShellViewPtr spView = NULL;
			if(SUCCEEDED(hr = spBrowser->QueryActiveShellView(&spView)))
			{
				//Could return directly the shell view, 
				//but since the caller wants a different interface in our case,
				//just help.
				
				if(SUCCEEDED(hr = spView->QueryInterface(riid, ppView)))
					hr = S_OK;
				else if(pErr != NULL)
				{
					if(hr == E_NOINTERFACE)
						*pErr = TEXT("IShellView does not support requested interface (likely IFolderView)");
					else
						*pErr = TEXT("IShellView::QueryInterface(out IFolderView) failed.");
				}
			}
			else if(pErr != NULL)
			{
				*pErr = TEXT("IShellBrowser::QueryActiveShellView(out IShellView) failed.");
			}
		}
		else if(pErr != NULL)
		{
			*pErr = TEXT("IServiceProvider::QueryService(SID_STopLevelBrowser, out IShellBrowser) failed.");
		}
	}
	else if(pErr != NULL)
	{
		if(hr == E_NOINTERFACE)
			*pErr = TEXT("Given IDispatch does not support IServiceProvider");
		else
			*pErr = TEXT("IDispatch::QueryInterface(out IServiceProvider) failed.");
	}
	
	return hr;
}

//Gets a the file path of a folder view's currently viewed folder.
//szPath be at least MAX_PATH characters in size.
EXTERN_C HRESULT GetFolderViewDirectoryPath(IFolderView *pView, LPTSTR szPath, LPCTSTR *pErr)
{
	HRESULT hrRet = E_UNEXPECTED;
	IPersistFolder2Ptr spPersist;
	if (SUCCEEDED( pView->GetFolder(IID_PPV_ARGS(&spPersist)) ))
	{
		PIDLIST_ABSOLUTE pidlFolder;
		if (SUCCEEDED(spPersist->GetCurFolder(&pidlFolder)))
		{
			if (SHGetPathFromIDList(pidlFolder, szPath))
			{
				hrRet = S_OK;
			}
			else if(pErr != NULL)
			{
				*pErr = TEXT("SHGetPathFromIDList(PCIDLIST_ABSOLUTE) failed.");
			}
			
			CoTaskMemFree(pidlFolder);
		}
		else if(pErr != NULL)
		{
			*pErr = TEXT("IPersistFolder2::GetCurFolder(out PIDLIST_ABSOLUTE) failed.");
		}
	}
	else if (pErr != NULL)
	{
		*pErr = TEXT("IFolderView::GetFolder(out IPersistFolder2) failed.");
	}
	return hrRet;
}

EXTERN_C HRESULT GetFolderViewDisplayName(IFolderView *pView, SHGDNF displayNameFlags, LPTSTR szName, UINT cchName, LPCTSTR *pErr)
{
	/*
	This code obtains the folder view's PIDL, 
	then binds to its parent folder, 
	then calls GetDisplayNameOf() and processes the returned STRRET.
	*/
	HRESULT hrRet = E_UNEXPECTED;
	IPersistFolder2Ptr spPersist;
	if (SUCCEEDED( pView->GetFolder(IID_PPV_ARGS(&spPersist)) ))
	{
		PIDLIST_ABSOLUTE pidlFolder;
		if (SUCCEEDED(spPersist->GetCurFolder(&pidlFolder)))
		{
			IShellFolderPtr spParentFolder;
			PCUITEMID_CHILD pidFolder;
			if (SUCCEEDED(SHBindToParent(pidlFolder, IID_PPV_ARGS(&spParentFolder), &pidFolder)))
			{
				STRRET str;
				if (SUCCEEDED(spParentFolder->GetDisplayNameOf(pidFolder, displayNameFlags, &str)))
				{
					if(SUCCEEDED(StrRetToBuf(&str, pidFolder, szName, cchName))) //Fred's note: This frees the STRRET's buffer
					{
						hrRet = S_OK;
					}
					else if(pErr != NULL)
					{
						*pErr = TEXT("StrRetToBuf(STRRET, PCUITEMID_CHILD, LPTSTR, UINT) failed.");
					}
				}
				else if(pErr != NULL)
				{
					*pErr = TEXT("IShellFolder::GetDisplayNameOf(PCUITEMID_CHILD, SHGDNF, out STRRET) failed.");
				}
			}
			else if(pErr != NULL)
			{
				*pErr = TEXT("SHBindToParent(PIDL, out IShellFolder, out PCUITEMID_CHILD) failed.");
			}
			
			CoTaskMemFree(pidlFolder);
		}
		else if(pErr != NULL)
		{
			*pErr = TEXT("IPersistFolder2::GetCurFolder(out PIDLIST_ABSOLUTE) failed.");
		}
	}
	else if (pErr != NULL)
	{
		*pErr = TEXT("IFolderView::GetFolder(out IPersistFolder2) failed.");
	}
	return hrRet;
}

/*
Result of this massive, waste-of-time test:
Folders don't implement IPersistMoniker, so one can guess other namespace extensions don't either.
*/
EXTERN_C HRESULT GetFolderViewMoniker(IFolderView *pView, LPTSTR szPath, LPCTSTR *pErr)
{
	HRESULT hrRet = E_UNEXPECTED;
	IPersistMonikerPtr spPersist;
	if (SUCCEEDED( pView->GetFolder(IID_PPV_ARGS(&spPersist)) ))
	{
		IMonikerPtr spMk;
		if (SUCCEEDED(spPersist->GetCurMoniker(&spMk)))
		{
			IBindCtxPtr spBindCtx;
			if(SUCCEEDED(CreateBindCtx(0, &spBindCtx)))
			{
				LPOLESTR displayName = NULL;
				if(SUCCEEDED(spMk->GetDisplayName(spBindCtx, NULL, &displayName)))
				{
					hrRet = S_OK;
					ATL::COLE2CT displayNameT(displayName);
					_tcsncpy_s(szPath, MAX_PATH, displayNameT, _TRUNCATE);
					CoTaskMemFree(displayName);
				}
				else if(pErr != NULL)
				{
					*pErr = TEXT("IMoniker::GetDisplayname(IBindCtx, NULL, out LPOLESTR) failed.");
				}
			}
			else if(pErr != NULL)
			{
				*pErr = TEXT("CreateBindCtx(out IBindCtx) failed.");
			}
		}
		else if(pErr != NULL)
		{
			*pErr = TEXT("IPersistFolder2::GetCurMoniker(out IMoniker) failed.");
		}
	}
	else if (pErr != NULL)
	{
		*pErr = TEXT("IFolderView::GetFolder(out IPersistMoniker) failed.");
	}
	return hrRet;
}

//Get the view's focused item's display name of requested type, if any
EXTERN_C HRESULT GetFolderViewFocusedItemDisplayName(IFolderView *pView, SHGDNF displayNameFlags, LPTSTR szName, UINT cchName)
{
	HRESULT hrRet = E_UNEXPECTED;
	int iFocus;
	//Get the view's focused item
	if (SUCCEEDED(pView->GetFocusedItem(&iFocus)))
	{
		PITEMID_CHILD pidItem;
		if (SUCCEEDED(pView->Item(iFocus, &pidItem)))
		{
			//Get the view's folder and call GetDisplayNameOf
			IShellFolderPtr spFolder;
			if (SUCCEEDED( pView->GetFolder(IID_PPV_ARGS(&spFolder)) ))
			{
				STRRET str;
				if (SUCCEEDED(spFolder->GetDisplayNameOf(pidItem, displayNameFlags, &str)))
				{
					if(SUCCEEDED(StrRetToBuf(&str, pidItem, szName, cchName))) //Fred's note: This frees the STRRET's buffer
					{
						hrRet = S_OK;
					}
				}
			}

			CoTaskMemFree(pidItem);
		}
	}
	return hrRet;
}