/*
Ceci est le programme "Querying information from an Explorer window"
	de Raymond Chen.
	http://blogs.msdn.com/oldnewthing/archive/2004/07/20/188696.aspx

Il a t quelque peu modifi pour afficher diffrents types de chemins.
Puis je l'ai modifi encore plus pour afficher toujours plus d'infos.
*/
#define STRICT
#define STRICT_TYPED_ITEMIDS
#include <windows.h>
#include <windowsx.h>
#include <ole2.h>
#include <commctrl.h>
#include <shlwapi.h>
#include "ShellCOMFunctions.h"
#include <stdio.h>
#include <tchar.h>

//Types
struct st_itemNames
{
	TCHAR Normal[MAX_PATH];
	TCHAR ParentRelativeParsing[MAX_PATH];
	TCHAR DesktopAbsoluteParsing[MAX_PATH];
	TCHAR ParentRelativeEditing[MAX_PATH];
	TCHAR DesktopAbsoluteEditing[MAX_PATH];
	//TCHAR FileSysPath[MAX_PATH]; = DesktopAbsoluteParsing
	//TCHAR URL[MAX_PATH]; = DesktopAbsoluteParsing
	TCHAR ParentRelativeForAddressbar[MAX_PATH];
	TCHAR ParentRelative[MAX_PATH];
};

typedef TCHAR filePath_t[MAX_PATH];

struct st_itemNameInfo
{
	filePath_t st_itemNames::* pPathBuffer;
	LPCTSTR pathName;
	SHGDNF flags;
};

//Constants
st_itemNameInfo const ITEM_NAME_INFOS[] = {
	{ &st_itemNames::Normal                     , TEXT("Normal")                      TEXT(":"), SHGDN_NORMAL },
	{ &st_itemNames::ParentRelative             , TEXT("ParentRelative")              TEXT(":"), SHGDN_INFOLDER },
	{ &st_itemNames::ParentRelativeForAddressbar, TEXT("ParentRelativeForAddressbar") TEXT(":"), SHGDN_INFOLDER|SHGDN_FORADDRESSBAR },
	{ &st_itemNames::DesktopAbsoluteParsing     , TEXT("DesktopAbsoluteParsing")      TEXT(":"), SHGDN_FORPARSING },
	{ &st_itemNames::ParentRelativeParsing      , TEXT("ParentRelativeParsing")       TEXT(":"), SHGDN_INFOLDER|SHGDN_FORPARSING },
	{ &st_itemNames::DesktopAbsoluteEditing     , TEXT("DesktopAbsoluteEditing")      TEXT(":"), SHGDN_FOREDITING },
	{ &st_itemNames::ParentRelativeEditing      , TEXT("ParentRelativeEditing")       TEXT(":"), SHGDN_INFOLDER|SHGDN_FOREDITING },
	{ NULL, NULL, 0 }
};
size_t const ITEM_NAME_INFOS_LENGTH = ARRAYSIZE(ITEM_NAME_INFOS) -1;

//Variables
HINSTANCE g_hinst;                          /* This application's HINSTANCE */
HWND g_hwndChild;                           /* Optional child window */

TCHAR g_szFolderPath[MAX_PATH];
struct st_itemNames g_folderNames;
struct st_itemNames g_focusedItemNames;
HWND g_lastShellWindow = NULL;

/*
 *	OnSize
 *		If we have an inner child, resize it to fit.
 */
void
OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	if(g_hwndChild) {
		MoveWindow(g_hwndChild, 0, 0, cx, cy, TRUE);
	}
}

void RecalcTextForDispatch(IDispatch* pDispWnd)
{
	LPCTSTR folderViewError = NULL;
	IFolderView *pView;
	HRESULT hr = GetShellWindowFolderView(pDispWnd, IID_IFolderView, (void**)&pView, &folderViewError);
	if(FAILED(hr))
	{
		_sntprintf_s(g_szFolderPath, MAX_PATH, _TRUNCATE, TEXT("<error in GetShellWindowFolderView: %s>"), folderViewError);
	}
	else
	{
		LPCTSTR folderPathError = NULL;
		hr = GetFolderViewDirectoryPath(pView, g_szFolderPath, &folderPathError);
		//hr = GetFolderViewMoniker(pView, g_szFolderPath, &folderPathError); //Test (failed)
		if(FAILED(hr))
		{
			_sntprintf_s(g_szFolderPath, MAX_PATH, _TRUNCATE, TEXT("<not a directory: %s>"), folderPathError);
		}
		
		for(size_t iInfo=0 ; iInfo<ITEM_NAME_INFOS_LENGTH ; iInfo++)
		{
			TCHAR *szNameBuffer = g_folderNames.*(ITEM_NAME_INFOS[iInfo].pPathBuffer);
			hr = GetFolderViewDisplayName(pView, ITEM_NAME_INFOS[iInfo].flags, szNameBuffer, MAX_PATH, &folderPathError);
			if(FAILED(hr))
				_sntprintf_s(szNameBuffer, MAX_PATH, _TRUNCATE, TEXT("<failed: %s>"), folderPathError);
		}
		
		for(size_t iInfo=0 ; iInfo<ITEM_NAME_INFOS_LENGTH ; iInfo++)
		{
			TCHAR *szNameBuffer = g_focusedItemNames.*(ITEM_NAME_INFOS[iInfo].pPathBuffer);
			GetFolderViewFocusedItemDisplayName(pView, ITEM_NAME_INFOS[iInfo].flags, szNameBuffer, MAX_PATH);
		}

		pView->Release();
	}
}

/* Note: I modified the function to use my reusable funcs. */
/* Note: The function will try to display the selected item's display names
	regardless of whether the current folder is a directory.*/
void CALLBACK RecalcText(HWND hwnd, UINT /*uMsg*/, UINT_PTR /*idEvent*/, DWORD /*dwTime*/)
{
	HWND hwndFind = GetForegroundWindow();
	g_szFolderPath[0] = TEXT('\0');

	for(size_t iInfo=0 ; iInfo<ITEM_NAME_INFOS_LENGTH ; iInfo++)
	{
		TCHAR *szNameBuffer = g_folderNames.*(ITEM_NAME_INFOS[iInfo].pPathBuffer);
		szNameBuffer[0] = TEXT('\0');
	}
	for(size_t iInfo=0 ; iInfo<ITEM_NAME_INFOS_LENGTH ; iInfo++)
	{
		TCHAR *szNameBuffer = g_focusedItemNames.*(ITEM_NAME_INFOS[iInfo].pPathBuffer);
		szNameBuffer[0] = TEXT('\0');
	}
	
	IDispatch *pDispWnd = NULL;
	HRESULT hr = FindShellWindowByHandle(hwndFind, &pDispWnd);
	if(FAILED(hr))
	{
		_sntprintf_s(g_szFolderPath, MAX_PATH, _TRUNCATE, TEXT("<error in FindShellWindow: %08lX>"), hr);
	}
	else if(hr==S_OK)
	{
		g_lastShellWindow = hwndFind;
		RecalcTextForDispatch(pDispWnd);
		pDispWnd->Release(), pDispWnd=NULL;
	}
	else
	{
		bool lastIsGood = false;
		if(g_lastShellWindow!=NULL && IsWindow(g_lastShellWindow))
		{
			hr = FindShellWindowByHandle(g_lastShellWindow, &pDispWnd);
			if(FAILED(hr))
			{
				_sntprintf_s(g_szFolderPath, MAX_PATH, _TRUNCATE, TEXT("<error in 2nd FindShellWindow: %08lX>"), hr);
			}
			else if(hr==S_OK)
			{
				lastIsGood = true;
				RecalcTextForDispatch(pDispWnd);
				pDispWnd->Release(), pDispWnd=NULL;
			}
			else { } //S_FALSE
		}
		else { } //S_FALSE

		if(!lastIsGood)
		{
			lstrcpyn(g_szFolderPath, TEXT("<no explorer window selected>"), MAX_PATH);
			g_lastShellWindow=NULL;
		}
	}
	InvalidateRect(hwnd, NULL, TRUE);
}

/*
 *	OnCreate
 *		Applications will typically override this and maybe even
 *		create a child window.
 */
BOOL
OnCreate(HWND hwnd, LPCREATESTRUCT lpcs)
{
	SetTimer(hwnd, 1, 10000, RecalcText);
	return TRUE;
}

/*
 *	OnDestroy
 *		Post a quit message because our application is over when the
 *		user closes this window.
 */
void
OnDestroy(HWND hwnd)
{
	PostQuitMessage(0);
}

/*
 *	PaintContent
 *		Interesting things will be painted here eventually.
 */
void
PaintContent(HWND hwnd, PAINTSTRUCT *pps)
{
	SIZE labelSize = { 0, 0 };
	LPCTSTR label1 = _T("Directory path: ");
	GetTextExtentPoint32(pps->hdc, label1, lstrlen(label1), &labelSize);
	TextOut(pps->hdc, 0, 0, label1, lstrlen(label1));
	TextOut(pps->hdc, labelSize.cx, 0, g_szFolderPath, lstrlen(g_szFolderPath));

	//Calculate the maximum size of a display name description.
	SIZE maxNameNameSize = { 0, 0 };
	for(size_t iInfo=0 ; iInfo<ITEM_NAME_INFOS_LENGTH ; iInfo++)
	{
		LPCTSTR sczNameName = ITEM_NAME_INFOS[iInfo].pathName;
		SIZE s = {0, 0};
		GetTextExtentPoint32(pps->hdc, sczNameName, lstrlen(sczNameName), &s);
		if(s.cx > maxNameNameSize.cx)
			maxNameNameSize.cx = s.cx;
		if(s.cy > maxNameNameSize.cy)
			maxNameNameSize.cy = s.cy;
	}

	const int FIRST_FOLDERNAME_Y = 50;
	//Space between the last folder name and the first focused item name
	const int DIFF_Y = 40;
	
	int yFirstFolderName = FIRST_FOLDERNAME_Y;
	int yFirstFocusedItemName = FIRST_FOLDERNAME_Y + ITEM_NAME_INFOS_LENGTH * maxNameNameSize.cy + DIFF_Y;

	LPCTSTR label2 = _T("Folder display names:");
	GetTextExtentPoint32(pps->hdc, label2, lstrlen(label2), &labelSize);
	TextOut(pps->hdc, 0, yFirstFolderName-labelSize.cy, label2, lstrlen(label2));
	
	for(size_t iInfo=0 ; iInfo<ITEM_NAME_INFOS_LENGTH ; iInfo++)
	{
		LPCTSTR sczNameName = ITEM_NAME_INFOS[iInfo].pathName;
		TCHAR *szNameBuffer = g_folderNames.*(ITEM_NAME_INFOS[iInfo].pPathBuffer);
		SIZE s = {0, 0};
		GetTextExtentPoint32(pps->hdc, sczNameName, lstrlen(sczNameName), &s);

		TextOut(pps->hdc, maxNameNameSize.cx - s.cx, (int)iInfo * maxNameNameSize.cy + yFirstFolderName, sczNameName , lstrlen(sczNameName));
		TextOut(pps->hdc, maxNameNameSize.cx + 10  , (int)iInfo * maxNameNameSize.cy + yFirstFolderName, szNameBuffer, lstrlen(szNameBuffer));
	}

	LPCTSTR label3 = _T("Focused item display names:");
	GetTextExtentPoint32(pps->hdc, label3, lstrlen(label3), &labelSize);
	TextOut(pps->hdc, 0, yFirstFocusedItemName-labelSize.cy, label3, lstrlen(label3));

	for(size_t iInfo=0 ; iInfo<ITEM_NAME_INFOS_LENGTH ; iInfo++)
	{
		LPCTSTR sczNameName = ITEM_NAME_INFOS[iInfo].pathName;
		TCHAR *szNameBuffer = g_focusedItemNames.*(ITEM_NAME_INFOS[iInfo].pPathBuffer);
		SIZE s = {0, 0};
		GetTextExtentPoint32(pps->hdc, sczNameName, lstrlen(sczNameName), &s);

		TextOut(pps->hdc, maxNameNameSize.cx - s.cx, (int)iInfo * maxNameNameSize.cy + yFirstFocusedItemName, sczNameName , lstrlen(sczNameName));
		TextOut(pps->hdc, maxNameNameSize.cx + 10  , (int)iInfo * maxNameNameSize.cy + yFirstFocusedItemName, szNameBuffer, lstrlen(szNameBuffer));
	}
}

/*
 *	OnPaint
 *		Paint the content as part of the paint cycle.
 */
void
OnPaint(HWND hwnd)
{
	PAINTSTRUCT ps;
	BeginPaint(hwnd, &ps);
	PaintContent(hwnd, &ps);
	EndPaint(hwnd, &ps);
}

/*
 *	OnPrintClient
 *		Paint the content as requested by USER.
 */
void
OnPrintClient(HWND hwnd, HDC hdc)
{
	PAINTSTRUCT ps = {0}; /*Init ajoute par Fred*/
	ps.hdc = hdc;
	GetClientRect(hwnd, &ps.rcPaint);
	PaintContent(hwnd, &ps);

}

/*
 *	Window procedure
 */
LRESULT CALLBACK
WndProc(HWND hwnd, UINT uiMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uiMsg) {

	HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
	HANDLE_MSG(hwnd, WM_SIZE, OnSize);
	HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
	HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
	case WM_PRINTCLIENT: OnPrintClient(hwnd, (HDC)wParam); return 0;
	}

	return DefWindowProc(hwnd, uiMsg, wParam, lParam);
}

BOOL
InitApp(void)
{
	WNDCLASS wc = {0}; /*Init ajoute par Fred*/

	wc.style = 0;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = g_hinst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = TEXT("Scratch"); /*Modif Fred:TCHAR*/

	if(!RegisterClass(&wc)) return FALSE;

	InitCommonControls();               /* In case we use a common control */

	return TRUE;
}


int WINAPI WinMain(
 HINSTANCE hinst, HINSTANCE hinstPrev,
 LPSTR lpCmdLine, int nShowCmd
 )
{
	MSG msg;
	HWND hwnd;

	g_hinst = hinst;

	if(!InitApp()) return 0;

	if(SUCCEEDED(CoInitialize(NULL))) {/* In case we use COM */

		hwnd = CreateWindow(
			TEXT("Scratch"),                /* Class Name (Modif Fred:TCHAR)*/
			TEXT("Scratch: ShellWindows1"), /* Title */
			WS_OVERLAPPEDWINDOW,            /* Style */
			CW_USEDEFAULT, CW_USEDEFAULT,   /* Position */
			CW_USEDEFAULT, CW_USEDEFAULT,   /* Size */
			NULL,                           /* Parent */
			NULL,                           /* No menu */
			hinst,                          /* Instance */
			0);                             /* No special parameters */

		ShowWindow(hwnd, nShowCmd);

		while(GetMessage(&msg, NULL, 0, 0)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		CoUninitialize();
	}

	return 0;
}
